from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import requests
from testcontainers.core.container import DockerContainer
from testcontainers.core.network import Network
from testcontainers.core.waiting_utils import wait_for_logs

from test.e2e.setup_testcontainers.artifacts import E2E_TESTDATA
from test.e2e.setup_testcontainers.artifacts import ArtifactDirs
from test.e2e.setup_testcontainers.ports import ExternalContainerPorts

FUSEKI_INTERNAL_PORT = 3030
SIPI_INTERNAL_PORT = 1024
INGEST_INTERNAL_PORT = 3340
API_INTERNAL_PORT = 3333


@dataclass(frozen=True)
class Containers:
    fuseki: DockerContainer
    sipi: DockerContainer
    ingest: DockerContainer
    api: DockerContainer


@dataclass(frozen=True)
class ContainerMetadata:
    artifact_dirs: ArtifactDirs
    versions: ImageVersions
    ports: ExternalContainerPorts
    names: ContainerNames


@dataclass(frozen=True)
class ImageVersions:
    fuseki: str
    sipi: str
    ingest: str
    api: str


@dataclass(frozen=True)
class ContainerNames:
    fuseki: str
    sipi: str
    ingest: str
    api: str


def get_all_containers(network: Network, metadata: ContainerMetadata) -> Containers:
    fuseki = _get_fuseki(network, metadata.versions.fuseki, metadata.ports, metadata.names)
    sipi = _get_sipi(network, metadata.versions.sipi, metadata.ports, metadata.names, metadata.artifact_dirs)
    ingest = _get_ingest(network, metadata.versions.ingest, metadata.ports, metadata.names, metadata.artifact_dirs)
    api = _get_api(network, metadata.versions.api, metadata.ports, metadata.names)
    containers = Containers(fuseki=fuseki, sipi=sipi, ingest=ingest, api=api)
    _print_containers_are_ready(containers)
    return containers


def _get_fuseki(
    network: Network, version: str, ports: ExternalContainerPorts, names: ContainerNames
) -> DockerContainer:
    fuseki = (
        DockerContainer(f"daschswiss/apache-jena-fuseki:{version}")
        .with_name(names.fuseki)
        .with_network(network)
        .with_bind_ports(host=ports.fuseki, container=FUSEKI_INTERNAL_PORT)
        .with_env("ADMIN_PASSWORD", "test")
    )
    fuseki.start()
    wait_for_logs(fuseki, f"Server .+ Started .+ on port {FUSEKI_INTERNAL_PORT}")
    print("Fuseki is ready")
    _create_data_set_and_admin_user(ports.fuseki)
    return fuseki


def _create_data_set_and_admin_user(fuseki_external_port: int) -> None:
    repo_config = Path("testdata/e2e/repo_config.ttl").read_text(encoding="utf-8")
    url = f"http://0.0.0.0:{fuseki_external_port}/$/datasets"
    files = {"file": ("file.ttl", repo_config, "text/turtle; charset=utf8")}
    if not requests.post(url, files=files, auth=("admin", "test"), timeout=30).ok:
        raise RuntimeError("Fuseki did not create the dataset")
    print("Dataset created")

    admin_user_data = Path("testdata/e2e/admin_user_data.ttl").read_text(encoding="utf-8")
    url = f"http://0.0.0.0:{fuseki_external_port}/knora-test/data?graph=http://www.knora.org/data/admin"
    files = {"file": ("file.ttl", admin_user_data, "text/turtle; charset: utf-8")}
    if not requests.post(url, files=files, auth=("admin", "test"), timeout=30).ok:
        raise RuntimeError("Fuseki did not create the admin user")
    print("Admin user created")


def _get_sipi(
    network: Network, version: str, ports: ExternalContainerPorts, names: ContainerNames, artifact_dirs: ArtifactDirs
) -> DockerContainer:
    sipi = (
        DockerContainer(f"daschswiss/knora-sipi:{version}")
        .with_name(names.sipi)
        .with_network(network)
        .with_bind_ports(host=ports.sipi, container=SIPI_INTERNAL_PORT)
        .with_env("KNORA_WEBAPI_KNORA_API_EXTERNAL_HOST", "0.0.0.0")  # noqa: S104
        .with_env("KNORA_WEBAPI_KNORA_API_EXTERNAL_PORT", ports.api)
        .with_command("--config=/sipi/config/sipi.docker-config.lua")
        .with_volume_mapping(artifact_dirs.tmp_sipi, "/tmp", "rw")  # noqa: S108
        .with_volume_mapping(E2E_TESTDATA, "/sipi/config", "rw")
        .with_volume_mapping(artifact_dirs.sipi_images, "/sipi/images", "rw")
    )
    sipi.start()
    wait_for_logs(sipi, f"Server listening on HTTP port {SIPI_INTERNAL_PORT}")
    print("Sipi is ready")
    return sipi


def _get_ingest(
    network: Network, version: str, ports: ExternalContainerPorts, names: ContainerNames, artifact_dirs: ArtifactDirs
) -> DockerContainer:
    ingest = (
        DockerContainer(f"daschswiss/dsp-ingest:{version}")
        .with_name(names.ingest)
        .with_network(network)
        .with_bind_ports(host=ports.ingest, container=INGEST_INTERNAL_PORT)
        .with_env("STORAGE_ASSET_DIR", "/opt/images")
        .with_env("STORAGE_TEMP_DIR", "/opt/tmp")
        # other containers are addressed with http://<service_name>:<internal_port>
        .with_env("JWT_ISSUER", f"http://{names.api}:{ports.api}")
        .with_env("JWT_SECRET", "UP 4888, nice 4-8-4 steam engine")
        .with_env("SIPI_USE_LOCAL_DEV", "false")
        .with_env("ALLOW_ERASE_PROJECTS", "true")
        .with_env("DB_JDBC_URL", "jdbc:sqlite:/opt/db/ingest.sqlite")
        .with_volume_mapping(artifact_dirs.sipi_images, "/opt/images", "rw")
        .with_volume_mapping(artifact_dirs.tmp_ingest, "/opt/tmp", "rw")
        .with_volume_mapping(artifact_dirs.ingest_db, "/opt/db", "rw")
    )
    ingest.start()
    wait_for_logs(ingest, "Started dsp-ingest")
    print("Ingest is ready")
    return ingest


def _get_api(network: Network, version: str, ports: ExternalContainerPorts, names: ContainerNames) -> DockerContainer:
    api = (
        DockerContainer(f"daschswiss/knora-api:{version}")
        .with_name(names.api)
        .with_network(network)
        # other containers are addressed with http://<service_name>:<internal_port>
        .with_env("KNORA_WEBAPI_DSP_INGEST_BASE_URL", f"http://{names.ingest}:{INGEST_INTERNAL_PORT}")
        .with_env("KNORA_WEBAPI_JWT_ISSUER", f"http://{names.api}:{ports.api}")
        .with_env("KNORA_WEBAPI_KNORA_API_EXTERNAL_PORT", ports.api)
        .with_env("KNORA_WEBAPI_TRIPLESTORE_HOST", names.fuseki)
        .with_env("KNORA_WEBAPI_TRIPLESTORE_FUSEKI_REPOSITORY_NAME", "knora-test")
        .with_env("KNORA_WEBAPI_TRIPLESTORE_FUSEKI_USERNAME", "admin")
        .with_env("KNORA_WEBAPI_TRIPLESTORE_FUSEKI_PASSWORD", "test")
        .with_env("ALLOW_ERASE_PROJECTS", "true")
        .with_bind_ports(host=ports.api, container=API_INTERNAL_PORT)
    )
    api.start()
    wait_for_logs(api, "AppState set to Running")
    wait_for_logs(api, "Starting api on")
    print("API is ready")
    return api


def _print_containers_are_ready(containers: Containers) -> None:
    print("Containers are ready")
    print(f"  {containers.fuseki._name}: {containers.fuseki.ports}")
    print(f"  {containers.sipi._name}: {containers.sipi.ports}")
    print(f"  {containers.ingest._name}: {containers.ingest.ports}")
    print(f"  {containers.api._name}: {containers.api.ports}")



import json
from pathlib import Path

import pytest
import regex

from dsp_tools.commands.excel2json.json_header import get_json_header
from dsp_tools.commands.excel2json.models.json_header import EmptyJsonHeader
from dsp_tools.commands.excel2json.models.json_header import FilledJsonHeader
from dsp_tools.commands.excel2json.models.json_header import JsonHeader
from dsp_tools.error.custom_warnings import DspToolsFutureWarning
from dsp_tools.error.exceptions import InputError


@pytest.fixture
def filled_json_header() -> JsonHeader:
    test_path = Path("testdata/excel2json/excel2json_files/json_header.xlsx")
    return get_json_header(test_path)


def test_get_json_header_no_file() -> None:
    test_path = Path("testdata/excel2json/old_excel2json_files")
    result = get_json_header(test_path)
    assert isinstance(result, EmptyJsonHeader)


def test_is_filled_header(filled_json_header: JsonHeader) -> None:
    assert isinstance(filled_json_header, FilledJsonHeader)


def test_serialised_header(filled_json_header: JsonHeader) -> None:
    serialised_header = filled_json_header.to_dict()
    with open("testdata/excel2json/expected_json_header.json", encoding="utf-8") as f:
        expected = json.load(f)
    assert expected == serialised_header


def test_to_dict_json_header_invalid_missing_sheet() -> None:
    test_path = Path("testdata/invalid-testdata/excel2json/json_header_missing_sheet.xlsx")
    expected = regex.escape(
        "The Excel file 'json_header.xlsx' contains the following problems:\n\n"
        "The following sheet(s) are mandatory and may not be empty:\n"
        "    - keywords"
    )
    with pytest.warns(
        DspToolsFutureWarning,
        match=regex.escape("The json_header.xlsx file does not have a sheet containing the enabled licenses"),
    ):
        with pytest.raises(InputError, match=expected):
            get_json_header(test_path)


def test_to_dict_json_header_invalid_empty_sheet() -> None:
    test_path = Path("testdata/invalid-testdata/excel2json/json_header_empty_sheet.xlsx")
    expected = regex.escape(
        "The Excel file 'json_header.xlsx' contains the following problems:\n\n"
        "The sheet 'description' has the following problems:\n\n"
        "At least one value is required in the columns: "
        "description_de, description_en, description_fr, description_it, description_rm\n"
        "Row 2 does not contain any values in those columns."
    )
    with pytest.warns(
        DspToolsFutureWarning,
        match=regex.escape("The json_header.xlsx file does not have a sheet containing the enabled licenses"),
    ):
        with pytest.raises(InputError, match=expected):
            get_json_header(test_path)


if __name__ == "__main__":
    pytest.main([__file__])



from dataclasses import dataclass


@dataclass
class XSDValidationMessage:
    line_number: int
    element: str | None
    attribute: str | None
    message: str


def get_xsd_validation_message_str(msg: XSDValidationMessage) -> str:
    msg_list = [f"Line Number {msg.line_number}"]
    if msg.element:
        msg_list.append(f"Element '{msg.element}'")
    if msg.attribute:
        msg_list.append(f"Attribute '{msg.attribute}'")
    msg_list.append(msg.message)
    return " | ".join(msg_list)



# mypy: disable-error-code="no-untyped-def"

from pathlib import Path

import pytest

from dsp_tools.cli.args import ValidateDataConfig
from dsp_tools.cli.args import ValidationSeverity
from dsp_tools.commands.validate_data.models.input_problems import DuplicateFileWarning
from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.input_problems import Severity
from dsp_tools.commands.validate_data.validation.check_duplicate_files import _get_filepaths_with_more_than_one_usage
from dsp_tools.commands.validate_data.validation.check_duplicate_files import check_for_duplicate_files
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValue
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValueMetadata
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource

TEST_ENV_CONFIG = ValidateDataConfig(
    xml_file=Path(),
    save_graph_dir=None,
    severity=ValidationSeverity.INFO,
    ignore_duplicate_files_warning=False,
    is_on_prod_server=False,
)

PROD_ENV_CONFIG = ValidateDataConfig(
    xml_file=Path(),
    save_graph_dir=None,
    severity=ValidationSeverity.INFO,
    ignore_duplicate_files_warning=False,
    is_on_prod_server=True,
)

FILEPATH_1 = "file_path_1.jpg"
FILEPATH_2 = "file_path_2.jpg"


@pytest.fixture
def file_value_1():
    return ParsedFileValue(FILEPATH_1, None, ParsedFileValueMetadata(None, None, None, None))


@pytest.fixture
def file_value_2():
    return ParsedFileValue(FILEPATH_2, None, ParsedFileValueMetadata(None, None, None, None))


class TestCheckDuplicates:
    def test_no_duplicates(self, file_value_1, file_value_2):
        resources = [
            ParsedResource("no_file", ":type", "lbl", None, [], None, None),
            ParsedResource("file_value_1", ":type", "lbl", None, [], file_value_1, None),
            ParsedResource("file_value_2", ":type", "lbl", None, [], file_value_2, None),
        ]
        result = check_for_duplicate_files(resources)
        assert not result

    def test_one_duplicate(self, file_value_1):
        resources = [
            ParsedResource("no_file", ":type", "lbl", None, [], None, None),
            ParsedResource("id_1", ":type", "lbl", None, [], file_value_1, None),
            ParsedResource("id_2", ":type", "lbl", None, [], file_value_1, None),
        ]
        result = check_for_duplicate_files(resources)
        assert isinstance(result, DuplicateFileWarning)
        assert len(result.problems) == 1
        problem = result.problems.pop(0)
        expected_msg = "value used 2 times"
        assert problem.problem_type == ProblemType.FILE_DUPLICATE
        assert not problem.res_id
        assert not problem.res_type
        assert problem.prop_name == "bitstream / iiif-uri"
        assert problem.severity == Severity.WARNING
        assert problem.message == expected_msg
        assert problem.input_value == file_value_1.value

    def test_several_duplicates(self, file_value_1, file_value_2):
        resources = [
            ParsedResource("no_file", ":type", "lbl", None, [], None, None),
            ParsedResource("file_value_1_1", ":type", "lbl", None, [], file_value_1, None),
            ParsedResource("file_value_1_2", ":type", "lbl", None, [], file_value_1, None),
            ParsedResource("file_value_1_3", ":type", "lbl", None, [], file_value_1, None),
            ParsedResource("file_value_2_1", ":type", "lbl", None, [], file_value_2, None),
            ParsedResource("file_value_2_2", ":type", "lbl", None, [], file_value_2, None),
        ]
        result = check_for_duplicate_files(resources)
        assert isinstance(result, DuplicateFileWarning)
        assert len(result.problems) == 2
        file_1 = next(x for x in result.problems if x.input_value == file_value_1.value)
        assert file_1.message == "value used 3 times"
        file_2 = next(x for x in result.problems if x.input_value == file_value_2.value)
        assert file_2.message == "value used 2 times"


class TestGetFilePathCountDict:
    def test_several_duplicates(self, file_value_1, file_value_2):
        file_1_number = 10
        file_2_number = 2
        resources = [
            ParsedResource(f"id_{i}", ":type", "lbl", None, [], file_value_1, None) for i in range(file_1_number)
        ]
        resources.extend(
            [ParsedResource(f"id_{i}", ":type", "lbl", None, [], file_value_2, None) for i in range(file_2_number)]
        )
        result = _get_filepaths_with_more_than_one_usage(resources)
        assert set(result.keys()) == {FILEPATH_1, FILEPATH_2}
        assert result[FILEPATH_1] == file_1_number
        assert result[FILEPATH_2] == file_2_number

    def test_one_duplicate(self, file_value_1, file_value_2):
        file_1_number = 10
        file_2_number = 1
        resources = [
            ParsedResource(f"id_{i}", ":type", "lbl", None, [], file_value_1, None) for i in range(file_1_number)
        ]
        resources.extend(
            [ParsedResource(f"id_{i}", ":type", "lbl", None, [], file_value_2, None) for i in range(file_2_number)]
        )
        result = _get_filepaths_with_more_than_one_usage(resources)
        assert set(result.keys()) == {FILEPATH_1}
        assert result[FILEPATH_1] == file_1_number

    def test_no_duplicate(self, file_value_1, file_value_2):
        resources = [
            ParsedResource("id_1", ":type", "lbl", None, [], file_value_1, None),
            ParsedResource("id_2", ":type", "lbl", None, [], file_value_2, None),
        ]
        result = _get_filepaths_with_more_than_one_usage(resources)
        assert not result



from collections import defaultdict

import pandas as pd

from dsp_tools.cli.args import ValidationSeverity
from dsp_tools.commands.validate_data.models.input_problems import AllProblems
from dsp_tools.commands.validate_data.models.input_problems import DuplicateFileWarning
from dsp_tools.commands.validate_data.models.input_problems import InputProblem
from dsp_tools.commands.validate_data.models.input_problems import MessageComponents
from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.input_problems import Severity
from dsp_tools.commands.validate_data.models.input_problems import SortedProblems
from dsp_tools.commands.validate_data.models.input_problems import UserPrintMessages

LIST_SEPARATOR = "\n    - "
GRAND_SEPARATOR = "\n\n----------------------------\n"


PROBLEM_TYPES_IGNORE_STR_ENUM_INFO = {ProblemType.GENERIC, ProblemType.FILE_VALUE_MISSING, ProblemType.FILE_DUPLICATE}


def sort_user_problems(
    all_problems: AllProblems, duplicate_file_warnings: DuplicateFileWarning | None
) -> SortedProblems:
    iris_removed, problems_with_iris = _separate_link_value_missing_if_reference_is_an_iri(all_problems.problems)
    filtered_problems = _filter_out_duplicate_problems(iris_removed)
    violations, warnings, info = _separate_according_to_severity(filtered_problems)
    if duplicate_file_warnings:
        warnings.extend(duplicate_file_warnings.problems)
    info.extend(problems_with_iris)
    unique_unexpected = list(set(x.component_type for x in all_problems.unexpected_results or []))
    return SortedProblems(
        unique_violations=violations,
        user_warnings=warnings,
        user_info=info,
        unexpected_shacl_validation_components=unique_unexpected,
    )


def _separate_according_to_severity(
    problems: list[InputProblem],
) -> tuple[list[InputProblem], list[InputProblem], list[InputProblem]]:
    violations = [x for x in problems if x.severity == Severity.VIOLATION]
    warnings = [x for x in problems if x.severity == Severity.WARNING]
    info = [x for x in problems if x.severity == Severity.INFO]
    return violations, warnings, info


def _separate_link_value_missing_if_reference_is_an_iri(
    problems: list[InputProblem],
) -> tuple[list[InputProblem], list[InputProblem]]:
    iris_referenced = []
    no_iris_referenced = []
    for prblm in problems:
        if prblm.problem_type != ProblemType.INEXISTENT_LINKED_RESOURCE:
            no_iris_referenced.append(prblm)
            continue
        if not prblm.input_value:
            no_iris_referenced.append(prblm)
        elif prblm.input_value.startswith("http://rdfh.ch/"):
            prblm.message = (
                "You used an absolute IRI to reference an existing resource in the DB. "
                "If this resource does not exist or is not of the correct type, an xmlupload will fail."
            )
            iris_referenced.append(prblm)
        else:
            no_iris_referenced.append(prblm)
    return no_iris_referenced, iris_referenced


def _filter_out_duplicate_problems(problems: list[InputProblem]) -> list[InputProblem]:
    grouped, without_res_id = _group_problems_by_resource(problems)
    filtered = without_res_id
    for problems_per_resource in grouped.values():
        text_value_filtered = _filter_out_duplicate_text_value_problem(problems_per_resource)
        file_value_corrected = _filter_out_duplicate_wrong_file_type_problems(text_value_filtered)
        filtered.extend(file_value_corrected)
    return filtered


def _filter_out_duplicate_text_value_problem(problems: list[InputProblem]) -> list[InputProblem]:
    filtered_problems = [x for x in problems if x.problem_type != ProblemType.VALUE_TYPE_MISMATCH]
    type_problems = [x for x in problems if x.problem_type == ProblemType.VALUE_TYPE_MISMATCH]

    grouped_dict = defaultdict(list)
    for prob in type_problems:
        grouped_dict[prob.prop_name].append(prob)

    for problem_list in grouped_dict.values():
        messages = [x.expected for x in problem_list]
        # Is there a chance of a duplicate (only possible for TextValue)?
        if "This property requires a TextValue" not in messages:
            filtered_problems.extend(problem_list)
            continue
        # Is there a more precise message about the type of TextValue?
        if "TextValue without formatting" not in messages and "TextValue with formatting" not in messages:
            # If there is not a more precise message, then the generic one is communicated to the user
            filtered_problems.extend(problem_list)
            continue
        # We remove the generic message and leave the specific one
        inx = messages.index("This property requires a TextValue")
        problem_list.pop(inx)
        filtered_problems.extend(problem_list)

    return filtered_problems


def _filter_out_duplicate_wrong_file_type_problems(problems: list[InputProblem]) -> list[InputProblem]:
    # If a class is for example, an AudioRepresentation, but a jpg file is used,
    # the created value is of type StillImageFileValue.
    # This creates a min cardinality (because the audio file is missing)
    # and a closed constraint violation (because it is not permissible to add an image)
    # However, we only want to give one message to the user
    idx_missing = next((i for i, x in enumerate(problems) if x.problem_type == ProblemType.FILE_VALUE_MISSING), None)
    idx_prohibited = next(
        (i for i, x in enumerate(problems) if x.problem_type == ProblemType.FILE_VALUE_PROHIBITED), None
    )
    if idx_missing is None or idx_prohibited is None:
        return problems
    missing_problem = problems[idx_missing]
    prohibited_problem = problems[idx_prohibited]
    # The result of the closed constraint violation, contains the input value,
    # while the message of the other shape is better, we want to include the actual input value.
    missing_problem.input_value = prohibited_problem.input_value
    return [problem for i, problem in enumerate(problems) if i not in {idx_missing, idx_prohibited}] + [missing_problem]


def _group_problems_by_resource(
    problems: list[InputProblem],
) -> tuple[dict[str, list[InputProblem]], list[InputProblem]]:
    grouped_res = defaultdict(list)
    problem_no_res_id = []
    for prob in problems:
        if not prob.res_id:
            problem_no_res_id.append(prob)
        else:
            grouped_res[prob.res_id].append(prob)
    return grouped_res, problem_no_res_id


def get_user_message(sorted_problems: SortedProblems, severity: ValidationSeverity) -> UserPrintMessages:
    """
    Creates the string to communicate the user message.

    Args:
        sorted_problems: validation problems
        severity: Severity level of validation information

    Returns:
        Problem message
    """
    violation_message, warning_message, info_message, unexpected_violations = None, None, None, None
    too_many_to_print = _are_there_too_many_to_print(sorted_problems, severity)
    if sorted_problems.unique_violations:
        if too_many_to_print:
            violation_body = None
            violation_df = _get_message_df(sorted_problems.unique_violations)
        else:
            violation_body = _get_problem_print_message(sorted_problems.unique_violations)
            violation_df = None
        violation_header = (
            f"During the validation of the data {len(sorted_problems.unique_violations)} errors were found. "
            f"Until they are resolved an xmlupload is not possible."
        )
        violation_message = MessageComponents(violation_header, violation_body, violation_df)
    if sorted_problems.user_warnings:
        if too_many_to_print:
            warning_body = None
            warning_df = _get_message_df(sorted_problems.user_warnings)
        else:
            warning_body = _get_problem_print_message(sorted_problems.user_warnings)
            warning_df = None
        warning_header = (
            f"During the validation of the data {len(sorted_problems.user_warnings)} "
            f"problems were found. Warnings are allowed on test servers. "
            f"Please note that an xmlupload on a prod sever will fail."
        )
        warning_message = MessageComponents(warning_header, warning_body, warning_df)
    if sorted_problems.user_info:
        if too_many_to_print:
            info_body = None
            info_df = _get_message_df(sorted_problems.user_info)
        else:
            info_body = _get_problem_print_message(sorted_problems.user_info)
            info_df = None
        info_header = (
            f"During the validation of the data {len(sorted_problems.user_info)} "
            f"potential problems were found. They will not impede an xmlupload."
        )
        info_message = MessageComponents(info_header, info_body, info_df)
    if sorted_problems.unexpected_shacl_validation_components:
        unexpected_header = "The following unknown violation types were found!"
        unexpected_body = LIST_SEPARATOR + LIST_SEPARATOR.join(sorted_problems.unexpected_shacl_validation_components)
        unexpected_violations = MessageComponents(unexpected_header, unexpected_body, None)
    return UserPrintMessages(violation_message, warning_message, info_message, unexpected_violations)


def _are_there_too_many_to_print(sorted_problems: SortedProblems, severity: ValidationSeverity) -> bool:
    number_of_problems = len(sorted_problems.unique_violations)
    if severity.value <= ValidationSeverity.WARNING.value:
        number_of_problems += len(sorted_problems.user_warnings)
    if severity.value == ValidationSeverity.INFO.value:
        number_of_problems += len(sorted_problems.user_info)
    return bool(number_of_problems > 60)


def _get_problem_print_message(problems: list[InputProblem]) -> str:
    grouped, without_res_id = _group_problems_by_resource(problems)
    messages = [_get_message_for_one_resource(without_res_id)] if without_res_id else []
    messages_with_ids = [
        _get_message_for_one_resource(v) for v in sorted(grouped.values(), key=lambda x: str(x[0].res_id))
    ]
    messages.extend(messages_with_ids)
    return GRAND_SEPARATOR.join(messages)


def _get_message_for_one_resource(problems: list[InputProblem]) -> str:
    if problems[0].res_id:
        start_msg = f"Resource ID: {problems[0].res_id} | Resource Type: {problems[0].res_type}"
    else:
        start_msg = ""
    prop_messages = _get_message_with_properties(problems)
    return f"{start_msg}\n{prop_messages}"


def _get_message_with_properties(problems: list[InputProblem]) -> str:
    messages = defaultdict(list)
    for prob in problems:
        messages[prob.prop_name].append(_get_message_detail_str(prob))

    def format_msg(propname: str, msg: list[str]) -> str:
        return f"{propname}{LIST_SEPARATOR}{LIST_SEPARATOR.join(msg)}"

    return "\n".join([format_msg(k, v) for k, v in messages.items()])


def _get_message_detail_str(problem: InputProblem) -> str:
    msg = []
    if problem.message:
        msg.append(problem.message)
    if problem.problem_type not in PROBLEM_TYPES_IGNORE_STR_ENUM_INFO:
        msg.append(str(problem.problem_type))
    if problem.input_value:
        msg.append(f"Your input: '{_shorten_input(problem.input_value, problem.problem_type)}'")
    if problem.input_type:
        msg.append(f"Actual input type: '{problem.input_type}'")
    if problem.expected:
        msg.append(f"Expected{_get_expected_prefix(problem.problem_type)}: {problem.expected}")
    return " | ".join(msg)


def _get_expected_prefix(problem_type: ProblemType) -> str | None:
    match problem_type:
        case ProblemType.VALUE_TYPE_MISMATCH:
            return " Value Type"
        case ProblemType.INPUT_REGEX:
            return " Input Format"
        case ProblemType.LINK_TARGET_TYPE_MISMATCH:
            return " Resource Type"
        case _:
            return ""


def _get_message_df(problems: list[InputProblem]) -> pd.DataFrame:
    problem_dicts = [_get_message_dict(x) for x in problems]
    df = pd.DataFrame.from_records(problem_dicts)
    df = df.sort_values(by=["Resource Type", "Resource ID", "Property"])
    return df


def _get_message_dict(problem: InputProblem) -> dict[str, str]:
    msg_dict = {
        "Resource ID": problem.res_id,
        "Resource Type": problem.res_type,
        "Property": problem.prop_name,
        "Your Input": _shorten_input(problem.input_value, problem.problem_type),
        "Input Type": problem.input_type,
    }
    non_empty_dict = {k: v for k, v in msg_dict.items() if v}
    expected_and_message = _get_expected_message_dict(problem)
    non_empty_dict.update(expected_and_message)
    if problem.problem_type not in PROBLEM_TYPES_IGNORE_STR_ENUM_INFO:
        non_empty_dict["Problem"] = str(problem.problem_type)
    return non_empty_dict


def _get_expected_message_dict(problem: InputProblem) -> dict[str, str]:
    out_dict = {}
    if problem.expected:
        msg_str = problem.expected
        if prefix := _get_expected_prefix(problem.problem_type):
            msg_str = f"{prefix.strip()}: {msg_str}"
        out_dict["Expected"] = msg_str
    if problem.message:
        if problem.expected:
            out_dict["Message"] = problem.message
        else:
            out_dict["Expected"] = problem.message
    return out_dict


def _shorten_input(user_input: str | None, problem_type: ProblemType) -> str | None:
    if problem_type in [
        ProblemType.FILE_DUPLICATE,
        ProblemType.FILE_VALUE_MISSING,
        ProblemType.FILE_VALUE_PROHIBITED,
        ProblemType.LINK_TARGET_TYPE_MISMATCH,
        ProblemType.INEXISTENT_LINKED_RESOURCE,
    ]:
        return user_input
    if not user_input:
        return None
    if user_input.startswith(("http://rdfh.ch/", " / http://rdfh.ch/lists/")):
        return user_input
    if len(user_input) < 51:
        return user_input
    return f"{user_input[:50]}[...]"



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from enum import auto
from pathlib import Path

from rdflib import Graph

from dsp_tools.utils.rdflib_constants import SubjectObjectTypeAlias


@dataclass
class SHACLGraphs:
    cardinality: Graph
    content: Graph


@dataclass
class RDFGraphs:
    data: Graph
    ontos: Graph
    cardinality_shapes: Graph
    content_shapes: Graph
    knora_api: Graph


@dataclass
class ValidationFilePaths:
    directory: Path
    data_file: str
    shacl_file: str
    report_file: str


@dataclass
class ValidationReportGraphs:
    conforms: bool
    validation_graph: Graph
    shacl_graph: Graph
    onto_graph: Graph
    data_graph: Graph


@dataclass
class QueryInfo:
    validation_bn: SubjectObjectTypeAlias
    focus_iri: SubjectObjectTypeAlias
    focus_rdf_type: SubjectObjectTypeAlias


@dataclass
class UnexpectedComponent:
    component_type: str


@dataclass
class ValidationResultBaseInfo:
    result_bn: SubjectObjectTypeAlias
    source_constraint_component: SubjectObjectTypeAlias
    focus_node_iri: SubjectObjectTypeAlias
    focus_node_type: SubjectObjectTypeAlias
    result_path: SubjectObjectTypeAlias
    severity: SubjectObjectTypeAlias
    detail: DetailBaseInfo | None = None


@dataclass
class DetailBaseInfo:
    detail_bn: SubjectObjectTypeAlias
    source_constraint_component: SubjectObjectTypeAlias


@dataclass
class ValidationResult:
    violation_type: ViolationType
    res_iri: SubjectObjectTypeAlias
    res_class: SubjectObjectTypeAlias
    severity: SubjectObjectTypeAlias
    property: SubjectObjectTypeAlias | None = None
    message: SubjectObjectTypeAlias | None = None
    input_value: SubjectObjectTypeAlias | None = None
    input_type: SubjectObjectTypeAlias | None = None
    expected: SubjectObjectTypeAlias | None = None


class ViolationType(Enum):
    SEQNUM_IS_PART_OF = auto()
    UNIQUE_VALUE = auto()
    VALUE_TYPE = auto()
    PATTERN = auto()
    GENERIC = auto()
    LINK_TARGET = auto()
    MAX_CARD = auto()
    MIN_CARD = auto()
    NON_EXISTING_CARD = auto()
    FILE_VALUE_PROHIBITED = auto()
    FILE_VALUE_MISSING = auto()


@dataclass
class ReformattedIRI:
    res_id: str
    res_type: str
    prop_name: str



import unittest

from dsp_tools.commands import excel2xml

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)


class TestListMappings(unittest.TestCase):
    def test_create_json_excel_list_mapping(self) -> None:
        # We start with an Excel column that contains list nodes, but with spelling errors
        excel_column = [
            "  first noude ov testlist  ",
            "sekond noude ov testlist",
            " fierst sobnode , sekond sobnode , Third Node Ov Testliest",  # multiple entries per cell are possible
            "completely wrong spelling variant of 'first subnode' that needs manual correction",
        ]
        corrections = {
            "completely wrong spelling variant of 'first subnode' that needs manual correction": "first subnode"
        }
        testlist_mapping_returned = excel2xml.create_json_excel_list_mapping(
            path_to_json="testdata/json-project/test-project-systematic.json",
            list_name="testlist",
            excel_values=excel_column,
            sep=",",
            corrections=corrections,
        )
        testlist_mapping_expected = {
            "first noude ov testlist": "first node of testlist",
            "sekond noude ov testlist": "second node of testlist",
            "fierst sobnode": "first subnode",
            "sekond sobnode": "second subnode",
            "Third Node Ov Testliest": "third node of testlist",
            "third node ov testliest": "third node of testlist",
            "completely wrong spelling variant of 'first subnode' that needs manual correction": "first subnode",
        }
        self.assertDictEqual(testlist_mapping_returned, testlist_mapping_expected)

    def test_create_json_list_mapping(self) -> None:
        testlist_mapping_returned = excel2xml.create_json_list_mapping(
            path_to_json="testdata/json-project/test-project-systematic.json",
            list_name="testlist",
            language_label="en",
        )
        testlist_mapping_expected = {
            "First node of the Test-List": "first node of testlist",
            "first node of the test-list": "first node of testlist",
            "First Sub-Node": "first subnode",
            "first sub-node": "first subnode",
            "Second Sub-Node": "second subnode",
            "second sub-node": "second subnode",
            "Second node of the Test-List": "second node of testlist",
            "second node of the test-list": "second node of testlist",
            "Third node of the Test-List": "third node of testlist",
            "third node of the test-list": "third node of testlist",
        }
        self.assertDictEqual(testlist_mapping_returned, testlist_mapping_expected)
